//--------------------------------------------------
// Basic procedures to programm SI4703 breakout
//
// Mostly based on 
//   - Sparkfun code
//   - Nathan Seidle code
//
// Some personal parts in here 
// or some modifications or corrections on my part...
// Frederic ABIGNOLI
//
//--------------------------------------------------
// Version 0.9
// 2020/06/27
//--------------------------------------------------

#define RESET_PIN 2		// reset pin for SI4703
#define SDIO_PIN A4		// SDA/A4 on Arduino
#define SCLK_PIN A5		// SCL/A5 on Arduino

//----------------------
// Init the SI4703 IC
// make it work with I2C
//----------------------
void si4703_init(void) {
	// We need to reset SI4703
	// and put in in I2C mode but making SDIO LOW 
	// and then reset LOW, delay and HIGH, delay
	pinMode(RESET_PIN, OUTPUT);
	pinMode(SDIO_PIN, OUTPUT);		// SDIO is connected to A4 for I2C
	digitalWrite(SDIO_PIN, LOW);	// A low SDIO indicates a 2-wire interface
	digitalWrite(RESET_PIN, LOW);	// Put Si4703 into reset
	delay(1);						// Some delays while we allow pins to settle
	digitalWrite(RESET_PIN, HIGH);	// Bring Si4703 out of reset with SDIO set to low and SEN pulled high with on-board resistor
	delay(1);						// Allow Si4703 to come out of reset
 
	Wire.begin(); 					// Init I2C
	
	// Set the oscillator
	si4703_readRegisters();				// Read the current register set
	si4703_registers[0x07] = 0x8100;	// Enable the oscillator, from AN230 page 9, rev 0.61 (works)
	si4703_updateRegisters();			// Update register
	delay(500);							// Wait for clock to settle - from AN230 page 9
	
	// Start the IC
	si4703_readRegisters();						//Read the current register set
	si4703_registers[POWERCFG] = 0x4001;		//Enable the IC
	
	// Start the RDS
	si4703_registers[SYSCONFIG1] |= (1<<RDS);	//Enable RDS
	
	// Spacing between station
#ifdef IN_EUROPE
	si4703_registers[SYSCONFIG1] |= (1<<DE);		//50kHz Europe setup
	si4703_registers[SYSCONFIG2] |= (1<<SPACE0);	//100kHz channel spacing for Europe
#else
	si4703_registers[SYSCONFIG2] &= ~(1<<SPACE1 | 1<<SPACE0) ; //Force 200kHz channel spacing for USA
#endif
	// Volume setting
	si4703_registers[SYSCONFIG2] &= 0xFFF0;			//Clear volume bits
	si4703_registers[SYSCONFIG2] |= 0x0001;			//Set volume to lowest
	
	// Mono reception
	si4703_registers[POWERCFG] |= (1 << MONOSELECT);	// Force mono reception
	
	si4703_updateRegisters(); //Update
 
 
	delay(110); //Max powerup time, from datasheet page 13
	memset(StationName,0,sizeof(StationName));				// reset current station name
	memset(StationLastName,0,sizeof(StationLastName));		// reset last station name decoded
	memset(RadiotextMess,0,sizeof(RadiotextMess));			// reset current radiotext
	memset(RadiotextLastMess,0,sizeof(RadiotextLastMess));	// reset last radiotext decoded
}
 
//-----------------------------------------------------
// Write the current 6 control registers (0x02 to 0x07)
// don't need to specify an address. The IC assume you
// are starting with 0x02 first...
//-----------------------------------------------------
byte si4703_updateRegisters(void) {
	Wire.beginTransmission(SI4703);
	for(int regSpot = 0x02 ; regSpot < 0x08 ; regSpot++) {
		byte high_byte = si4703_registers[regSpot] >> 8;
		byte low_byte = si4703_registers[regSpot] & 0x00FF;
 
		Wire.write(high_byte); //Upper 8 bits
		Wire.write(low_byte); //Lower 8 bits
	}
 
	//End this transmission
	byte ack = Wire.endTransmission();
	if(ack != 0) { //We have a problem! 
		// Serial.print("Write Fail:"); //No ACK!
		// Serial.println(ack, DEC); //I2C error: 0 = success, 1 = data too long, 2 = rx NACK on address, 3 = rx NACK on data, 4 = other error
		return(FAIL);
	}
	return(SUCCESS);
}

//----------------------------------------
// Dump current SI4703 registers on serial
//----------------------------------------
void si4703_printRegisters(void) {
	si4703_readRegisters();				// Read all the registers
	for(int x = 0 ; x < 16 ; x++) {		// Print the 16 registers
		sprintf(printBuffer, "Reg 0x%02X = 0x%04X", x, si4703_registers[x]);
		Serial.println(printBuffer);
  }
}

//-------------------------------------------------------
// read SI4703 internal register
// All the control set from 0x00 to 0x0F will be returned
//-------------------------------------------------------
void si4703_readRegisters(void){
	//Si4703 begins reading from register upper register of 0x0A and reads to 0x0F, then loops to 0x00.
	Wire.requestFrom(SI4703, 32); //We want to read the entire register set from 0x0A to 0x09 = 32 bytes.
 
	while(Wire.available() < 32) ; //Wait for 16 words/32 bytes to come back from slave I2C device
	//We may want some time-out error here
 
	//Remember, register 0x0A comes in first so we have to shuffle the array around a bit
	for(int x = 0x0A ; ; x++) { //Read in these 32 bytes
		if(x == 0x10)
			x = 0; //Loop back to zero
		si4703_registers[x] = Wire.read() << 8;
		si4703_registers[x] |= Wire.read();
		if(x == 0x09) 
			break; //We're done!
	}
}
 


//--------------------------
// Return current channel
//--------------------------
int readChannel(void) {
	si4703_readRegisters();
	int channel = si4703_registers[READCHAN] & 0x03FF; //Mask out everything but the lower 10 bits
#ifdef IN_EUROPE
	channel *= 1; //98 * 1 = 98
#else
	channel *= 2; //49 * 2 = 98
#endif
	channel += 875; //98 + 875 = 973
	return(channel);
}

//-------------------------
// Reglage du volume global
//-------------------------
void setVolume(byte myvolume) {
	// If level not in [0..16] return
	if (myvolume<=0 || myvolume  >=16)
		return;
	si4703_registers[SYSCONFIG2] &= 0xFFF0; 	//Clear volume bits
	si4703_registers[SYSCONFIG2] |= myvolume; 	//Set new volume
	si4703_updateRegisters(); 					//Update
}

//------------------------
// Reglage de la frequence
//------------------------
void gotoChannel(int ChannelTuning){
	StationNameStatus=0;		// Reset Station name status
	RadiotextStatus=0;			// Reset Radiotext message status

	ChannelTuning *= 10; 		//973 * 10 = 9730
	ChannelTuning -= 8750;		//9730 - 8750 = 980
#ifdef IN_EUROPE
	// 100khz between station
	ChannelTuning /= 10;	//980 / 10 = 98
#else
	// 200khz between station
	ChannelTuning /= 20;	//980 / 20 = 49
#endif
 
	//These steps come from AN230 page 20 rev 0.5
	si4703_readRegisters();
	si4703_registers[CHANNEL] &= 0xFE00; //Clear out the channel bits
	si4703_registers[CHANNEL] |= ChannelTuning; //Mask in the new channel
	si4703_registers[CHANNEL] |= (1<<TUNE); //Set the TUNE bit to start
	si4703_updateRegisters();
 
	// Verify flag STC
	// Need to be one to validate tuning
	while(1) {
		si4703_readRegisters();
		if ( (si4703_registers[STATUSRSSI] & (1<<STC)) != 0 ) 
			break; //Tuning complete!
	}
	// Refresh local registers
	si4703_readRegisters();
	// Clear tune bit
	si4703_registers[CHANNEL] &= ~(1<<TUNE);
	// Update registers
	si4703_updateRegisters();
 
	// Verify flag STC
	// Need to be one to validate tuning
	while(1) {
		si4703_readRegisters();
		if( (si4703_registers[STATUSRSSI] & (1<<STC)) == 0) 
			break; //Tuning complete!
	}
}

//------------------------------------------
// Read station name
// Messages from group 0A ou 0B (BASIC TUNING)
// Based on Sparkfun Code
//-------------------------------------------
void readRdsStation(char* buffer, long timeout) { 
long endTime = millis() + timeout;
	while(completedCountSN  < SN_PARTS && millis() < endTime) {
		si4703_readRegisters();
		if(si4703_registers[STATUSRSSI] & (1<<RDSR)){
			// ls 2 bits of B determine the 4 letter pairs
			// once we have a full set return
			// if you get nothing after 20 readings return with empty string
			blockB = si4703_registers[RDSB];		// Retrieve Bloc B content
			GroupType = blockB >> 12;  				// Retrieve Group Type Number
			if (GroupType == MESS_BASIC_TUNING) {	// If group type BASIC_TUNING
				// Serial.print(blockB,HEX);			// retr
				indexSN = blockB & 0x03;				// Retrieve index
				// Serial.print(":(");
				// Serial.print(index);
				// Serial.print(")");
				blockD = si4703_registers[RDSD];	// retrieve BlockD content
				// Serial.print(":");
				// Serial.println(blockD,HEX);
		  
				if (! completedSName[indexSN]) {
					completedSName[indexSN] = true;
					completedCountSN ++;
					char Dh = (blockD & 0xFF00) >> 8;
					char Dl = (blockD & 0x00FF);
					buffer[indexSN * 2] = Dh;
					buffer[indexSN * 2 +1] = Dl;
					// Serial.print(si4703_registers[RDSD]); Serial.print(" ");
					// Serial.print(index);Serial.print(" ");
					// Serial.write(Dh);
					// Serial.write(Dl);
					// Serial.println();
				}
			}
			delay(40); //Wait for the RDS bit to clear
		}
		else {
			delay(30); //From AN230, using the polling method 40ms should be sufficient amount of time between checks
		}
	}
	if (millis() >= endTime) {
		buffer[0] ='\0';
		return;
	}

	buffer[8] = '\0';
}

//--------------------------------------
// Rds Station Name reading
// Non blocking proc 
// Content of StationNameStatus
//    0 : reset buffer
//    1 : Station name ready to be decoded
//    2 : Station named decoded
//    3 : Station named decoded and read
//   10 : Error
// Update the address of buffer with the content
// Do a rds request only if interval between 
// two request is reached 
//--------------------------------------
void readRdsStationNB(char* buffer, int interval) { 
static long LastDecodedTimeSN=0;		// Time of the last request
	// Verify if job has already been done
	if (StationNameStatus >= SN_READY) {				// if name already decoded or read or error
		return;											// return
	}

	// Check if Station Name need to be cleared
	if (StationNameStatus==SN_RESET) {					// If reset required
		// Serial.println("Reset requested...");
		memset(StationName,0,sizeof(StationName));		// Clear buffer
		StationNameStatus=SN_DECODING;					// ready to decode the next name
		completedCountSN=0;								// Reset number of decoded parts
		StationNameCycles=0;							// reset decoder cycles
		LastDecodedTimeSN=millis();						// reset time of the last decode
		for (int i=0; i<SN_PARTS; i++)
			completedSName[i]=false;
		return;											// return
	}

	if (StationNameCycles >= MAX_SN_CYCLES ) {			// if maximum number of decoder cycles is reached	
		// Serial.println("Max Cycles reached...");
		StationNameStatus=SN_ERROR;						// status error
		return;											// return
	}
	
	// If we are here, we have to decode the name
	
	// We check if it's not to early
	if ( (LastDecodedTimeSN + interval) > millis() )		// Check if it's time to send RDS Request
		return;											// If not then return
	else
		LastDecodedTimeSN=millis();						// reset LastDecodedTime for next decode
		
	StationNameCycles++;								// new decoder cycle
	si4703_readRegisters();								// read all the SI4703 internal registers
	if(si4703_registers[STATUSRSSI] & (1<<RDSR)){		// If RDSR is 1, we have something to decode
	Serial.print(".");	
		blockB = si4703_registers[RDSB];				// Retrieve Bloc B content
		GroupType = blockB >> 12;  						// Retrieve Group Type Number
		if (GroupType != MESS_BASIC_TUNING)				// If group is not 0A or 0B
			return;										// return
		indexSN = blockB & 0x03;						// Retrieve index of the Station name part cuurently received
		blockD = si4703_registers[RDSD];				// retrieve BlockD content (2 char of station name)
		if (! completedSName[indexSN]) {
			completedSName[indexSN] = true;
			completedCountSN ++;
			char Dh = (blockD & 0xFF00) >> 8;
			char Dl = (blockD & 0x00FF);
			buffer[indexSN * 2] = Dh;
			buffer[indexSN * 2 +1] = Dl;
		}
		// having 4 parts decoded doesn't mean that the content is so great
		// I decide to do a minimum of 2 decoding
		// and if the content is equal the 2 time
		// perhaps the content "should be" OK...
		if (completedCountSN >=SN_PARTS ) {										// If the 4 parts are decoded
			if (!memcmp(StationName,StationLastName,sizeof(StationName))) {		// compare with last received
				StationNameStatus=SN_READY;										// Station name OK
				return;
			}
			else {																// different, let's for another cycle
				memcpy(StationLastName,StationName,sizeof(StationName));		// it should do this until Stationname == StationLastName
				for (int i=0;i<SN_PARTS;i++)
					completedSName[i]=false;
				completedCountSN=0;
				StationNameCycles=0;
			}
			return;
		}
	}
}

//------------------------------------------------------------
// Reset radiotext context
// clear all the internal variables needed to decode radiotext
//------------------------------------------------------------
void ResetRadiotextContext(byte part) {
	memset(RadiotextMess,0,sizeof(RadiotextMess));	// Clear Current radiotext
	if (part==RT_RST_FULL)
		memset(RadiotextLastMess,0,sizeof(RadiotextLastMess));	// Clear Last Radiotext
	//RadiotextStatus=RT_DECODING;					// ready to decode the next name
	completedCountRT=0;								// Reset number of decoded parts
	RadiotextCycles=0;								// reset decoder cycles
	for (int i=0; i < RT_PARTS; i++)
		completedRadiotext[i]=false;
}

//------------------------------------------------------------
// Reset radiotext context
// clear all the internal variables needed to decode radiotext
//------------------------------------------------------------
void ResetStationNameContext(void) {
	memset(StationName,0,sizeof(StationName));		// Clear buffer
	StationNameStatus=SN_DECODING;					// ready to decode the next name
	completedCountSN=0;								// Reset number of decoded parts
	StationNameCycles=0;						// reset decoder cycles
	for (int i=0; i < SN_PARTS; i++)
		completedSName[i]=false;
}
//--------------------------------------
// Radiotext Mess Reading
// Non blocking proc
// Content of RadiotextStatus
//    0 : reset buffer
//    1 : Station name ready to be decoded
//    2 : Station named decoded
//    3 : Station named decoded and read
//   10 : Error
// Update the address of buffer with the content
// Do a rds request only if interval between 
// two request is reached 
//--------------------------------------
void readRdsRadiotextNB(char* buffer, int interval) { 
static long LastDecodedTimeRT=0;		// Time of the last request
char Ch, Cl, Dh, Dl;
	if (RadiotextCycles >= MAX_RT_CYCLES ) 			// if maximum number of decoder cycles is reached	
		return;
	// Verify if job has already been done
	// if (RadiotextStatus >= RT_READY) {				// if name already decoded or read or error
		// return;											// return
	// }

	// // Check if Radio text need to be cleared
	// if (RadiotextStatus==RT_RESET) {					// If reset required
		// Serial.println("Reset requested...");
		// ResetRadiotextContext();
		// LastDecodedTimeRT=millis();						// reset time of the last decode
		// return;											// return
	// }

	// if (RadiotextCycles >= MAX_RT_CYCLES ) {			// if maximum number of decoder cycles is reached	
		 // // Serial.println("Max Cycles reached...");
		// RadiotextStatus=RT_ERROR;						// status error
		// return;											// return
	// }
	
	// If we are here, we have to decode the name
	
	// We check if it's not to early
	if ( (LastDecodedTimeRT + interval) > millis() )		// Check if it's time to send RDS Request
		return;											// If not then return
	else
		LastDecodedTimeRT=millis();						// reset LastDecodedTimeRT for next decode
		
	RadiotextCycles++;									// new decoder cycle
	si4703_readRegisters();								// read all the SI4703 internal registers
	if(si4703_registers[STATUSRSSI] & (1<<RDSR)){		// If RDSR is 1, we have somethin to decode
		blockB = si4703_registers[RDSB];				// Retrieve Bloc B content
		GroupType = blockB >> 12;  						// Retrieve Group Type Number
		if (GroupType != MESS_RADIO_TEXT)				// If group is not 2A or 2B
			return;										// return
		indexRT = blockB & 0x0F;						// Retrieve index of the radiotext part cuurently received
		if (indexRT == 0) {								// If index==0 then we have a new message
			ResetRadiotextContext(RT_RST_PARTIAL);		// Reset radiotext context
			LastDecodedTimeRT=millis();					// reset time of the last decode
		}
		blockC = si4703_registers[RDSC];				// retrieve BlockC content (2 first char of radiotext)
		blockD = si4703_registers[RDSD];				// retrieve BlockD content (2 last char of radiotext)
		if (! completedRadiotext[indexRT]) {
			completedRadiotext[indexRT] = true;
			completedCountRT ++;
			Dh = (blockD & 0xFF00) >> 8;
			Dl = (blockD & 0x00FF);
			Ch = (blockC & 0xFF00) >> 8;
			Cl = (blockC & 0x00FF);
			buffer[indexRT * 4] = Ch;
			buffer[indexRT * 4 + 1] = Cl;
			buffer[indexRT * 4 + 2] = Dh;
			buffer[indexRT * 4 + 3] = Dl;

		}
		if (completedCountRT >= RT_PARTS || Ch==13 || Cl==13 || Dh==13 || Dl==13) {	//  
			if (memcmp(RadiotextMess,RadiotextLastMess,4 * RT_PARTS)) {		// Compare if current message is different of the last received
				RadiotextStatus=RT_READY;									// If it's the case then RT_READY
				memcpy(RadiotextLastMess,RadiotextMess,4 * RT_PARTS);		// Copy current message in last message buffer
			}
			return;
		}
	}
}

//---------------------------------------
// Seeks out the next available station
// Returns the freq if it made it
// Returns zero if failed
//---------------------------------------
byte seek(byte seekDirection) {
	si4703_readRegisters();							// Read all registers
 
	//Set seek mode wrap bit
	//si4703_registers[POWERCFG] |= (1<<SKMODE);	//Allow wrap
	si4703_registers[POWERCFG] &= ~(1<<SKMODE); 	// Disallow wrap - if you disallow wrap, you may want to tune to 87.5 first
 
	if(seekDirection == SEEK_DOWN) 
		si4703_registers[POWERCFG] &= ~(1<<SEEKUP);	// Seek down is the default upon reset
	else
		si4703_registers[POWERCFG] |= 1<<SEEKUP;	// Set the bit to seek up
 
	si4703_registers[POWERCFG] |= (1<<SEEK);		// Start seek
 
	si4703_updateRegisters();						// Seeking will now start
 
	// Poll to see if STC is set
	while(1) {
		si4703_readRegisters();
		if((si4703_registers[STATUSRSSI] & (1<<STC)) != 0) 
			break; //Tuning complete!
 
		// Serial.print("Trying station:");
		// Serial.println(readChannel());
	}
 
	si4703_readRegisters();
	int valueSFBL = si4703_registers[STATUSRSSI] & (1<<SFBL); //Store the value of SFBL
	si4703_registers[POWERCFG] &= ~(1<<SEEK); //Clear the seek bit after seek has completed
	si4703_updateRegisters();
	
	//Wait for the si4703 to clear the STC as well
	while(1) {
		si4703_readRegisters();
		if( (si4703_registers[STATUSRSSI] & (1<<STC)) == 0)
			break; //Tuning complete!
		// Serial.println("Waiting...");
	}
 
	if(valueSFBL) { //The bit was set indicating we hit a band limit or failed to find a station
		Serial.println("Seek limit hit");	//Hit limit of band during seek
		return(FAIL);
	}
	
	// Serial.println("Seek complete");		//Tuning complete!
	StationNameStatus=0;					// to reset Station Name or Radiotext buffer
	return(SUCCESS);

}



void debugRDS(void) {
    if(KeyboardInput == '0') {
      Serial.println("Poll RDS - x to exit");
      while(1) {
        if(Serial.available() > 0)
          if(Serial.read() == 'x') break;
 
        si4703_readRegisters();
        if(si4703_registers[STATUSRSSI] & (1<<RDSR)){
          //Serial.println("We have RDS!");
          byte Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl;
          Ah = (si4703_registers[RDSA] & 0xFF00) >> 8;
          Al = (si4703_registers[RDSA] & 0x00FF);
 
          Bh = (si4703_registers[RDSB] & 0xFF00) >> 8;
          Bl = (si4703_registers[RDSB] & 0x00FF);
 
          Ch = (si4703_registers[RDSC] & 0xFF00) >> 8;
          Cl = (si4703_registers[RDSC] & 0x00FF);
 
          Dh = (si4703_registers[RDSD] & 0xFF00) >> 8;
          Dl = (si4703_registers[RDSD] & 0x00FF);
 
		  tempbyte=Bh>>4;

		Serial.print("RDS: ");
         Serial.print(Ah,HEX);
		  Serial.print(":");
          Serial.print(Al,HEX);
		  Serial.print(":");
		  Serial.print("(");
		  Serial.print(tempbyte);
		  Serial.print(")");
          Serial.print(Bh,HEX);
		  Serial.print(":");
          Serial.print(Bl,HEX);
		  Serial.print(":");
          Serial.print(Ch,HEX);
		  Serial.print(":");
          Serial.print(Cl,HEX);
		  Serial.print(":");
          Serial.print(Dh,HEX);
		  Serial.print(":");
          Serial.print(Dl,HEX);

		  char outputstr[10];
		// decode Station ID
		if (tempbyte==0) {
				Serial.print("[");
				 sprintf(printBuffer, "%c%c", Dh,Dl);
				Serial.print(printBuffer);
				Serial.print("]");
				
		}
	
		// Decode Radiotext
		if (tempbyte==2) {
				if (Bl==0 )
					Serial.println();
				// Serial.print("[");
				 sprintf(printBuffer, "%c%c%c%c", Ch,Cl,Dh,Dl);
				Serial.print(printBuffer);
				Serial.print("(");
          Serial.print(Ch,HEX);
		  Serial.print(":");
          Serial.print(Cl,HEX);
		  Serial.print(":");
          Serial.print(Dh,HEX);
		  Serial.print(":");
          Serial.print(Dl,HEX);
				Serial.print(")");
				// Serial.print("]");
				
				
		}

		  Serial.println(" !");
          delay(40); //Wait for the RDS bit to clear
        }
        else {
          //Serial.println("No RDS");
          delay(30); //From AN230, using the polling method 40ms should be sufficient amount of time between checks
        }
        //delay(500);
      }
    }

}

//----------------------------------------
// Change current channel after some delay
//----------------------------------------
void changeFrequencyOnDelay(void) {
	if (timeFreqChange) {
		if (timeFreqChange+DELAY_CHANNEL_CHANGE < millis()) {		// If time elapsed
			gotoChannel(currentChannel);							// change current channel on the tuner
			lcd.setCursor(11,0);
			lcd.print(F("<No RDS>"));
			timeFreqChange=0;										// no change to do anymore (until real change with encoders)
		}
	}
}

